#pragma once
#include <fstream>
#include <iostream>
#include <source_location>
#include <jlog/Token.hpp>
#include <jlog/Timestamp.hpp>
#include <Event.h>

namespace jlog {
    class Logger {
    public:

        Logger(std::ostream &stream = std::cout) : os{stream} {};
    public:
        void operator()(const std::string &message, const std::source_location &location = std::source_location::current(),
                   const Timestamp &ts = Timestamp()) {
            if (Enable)
                Log(message, location, ts);
        };
    protected:
        virtual void
        Log(const std::string &message, const std::source_location &location = std::source_location::current(),
            const Timestamp &ts = Timestamp()) = 0;

    public:
        bool Enable = true;
        bool IncludeLocation = true;
        bool IncludeTimestamp = true;
    protected:
        std::ostream &os = std::cout;
    };

    class ColorLogger : public Logger {
    public:
        ColorLogger(std::ostream &stream = std::cout) : Logger(stream) {};
    public:
        bool IncludeColor = true;
    };

    class ConsoleLogger : public ColorLogger {
    public:
        //explicit ConsoleLogger();
        explicit ConsoleLogger(const std::string &context) : ColorLogger() { this->context = context; };
    protected:
        virtual void
        Log(const std::string &message, const std::source_location &location = std::source_location::current(),
            const Timestamp &ts = Timestamp()) {
            if (!Enable)
                return;
            // Disable color on Windows. I've been trying my best to keep this macro hidden from the user. - maxine
#ifdef WIN32
            IncludeColor = false;
#endif
            if (IncludeTimestamp) {
                Timestamp ts;
                os << token{std::format(
                        "{}-{}-{} {}:{}:{}.{}",
                        ts.Year(),
                        ts.Month(),
                        ts.Day(),
                        ts.Hour().count(),
                        ts.Minute().count(),
                        ts.Second().count(),
                        ts.Millisecond().count()), "[]", timestampColor}.Stringer(IncludeColor);
            }

            os << token(context, "[]", contextColor).Stringer(IncludeColor);

            if (IncludeLocation) {
                os << token{location.function_name(), "[]", locationColor}.Stringer(IncludeColor);
                os << token{std::format("{}:{}", location.file_name(), location.line()), "[]", locationColor}.Stringer(
                        IncludeColor);
            }

            os << token{">>", "", pointerColor}.Stringer(IncludeColor)
               << token(message, "", messageColor).Stringer(IncludeColor) << "\n";
        }

    protected:
        std::string context = "CONSOLE";
    protected:
        Color4 timestampColor = Colors::Purples::Fuchsia;
        Color4 contextColor = Colors::White;
        Color4 locationColor = Colors::Pinks::Pink;
        Color4 pointerColor = Colors::Pinks::LightPink;
        Color4 messageColor = Colors::Greens::LightGreen;
    };

    class FileLogger : public Logger {
    public:
        //explicit FileLogger();
        explicit FileLogger(const std::string &context, std::ofstream &file) : Logger(
                file) { this->context = context; };
    protected:
        virtual void
        Log(const std::string &message, const std::source_location &location = std::source_location::current(),
            const Timestamp &ts = Timestamp()) {
            if (!Enable)
                return;

            if (IncludeTimestamp) {
                Timestamp ts;
                os << token{std::format(
                        "{}-{}-{} {}:{}:{}.{}",
                        ts.Year(),
                        ts.Month(),
                        ts.Day(),
                        ts.Hour().count(),
                        ts.Minute().count(),
                        ts.Second().count(),
                        ts.Millisecond().count()), "[]"}.Stringer(false);
            }

            os << token(context, "[]").Stringer(false);

            if (IncludeLocation) {
                os << token{location.function_name(), "[]"}.Stringer(false);
                os << token{std::format("{}:{}", location.file_name(), location.line()), "[]"}.Stringer(false);
            }

            os << token{">>", ""}.Stringer(false) << token(message, "").Stringer(false) << "\n";
        }

    protected:
        std::string context = "FILE";
    };

    class CompoundLogger : protected ConsoleLogger, protected FileLogger {
    public:
        Event<std::string, Color4> OnLog;
        explicit CompoundLogger(const std::string &context, std::ofstream &file) : ConsoleLogger(context),
                                                                                   FileLogger(context, file) {};
    public:
        void
        operator()(const std::string &message, const std::source_location &location = std::source_location::current(),
                   const Timestamp &ts = Timestamp()) { Log(message, location, ts); };

        virtual void
        Log(const std::string &message, const std::source_location &location = std::source_location::current(),
            const Timestamp &ts = Timestamp()) {
            OnLog.Invoke(message, messageColor);
            ConsoleLogger::Log(message, location, ts);
            FileLogger::Log(message, location, ts);
        }

    public:
        void EnableConsole(bool b) { ConsoleLogger::Enable = b; };

        void EnableFile(bool b) { FileLogger::Enable = b; };

        void IncludeTimestamp(bool b) {
            ConsoleLogger::IncludeTimestamp = b;
            FileLogger::IncludeTimestamp = b;
        };

        void IncludeLocation(bool b) {
            ConsoleLogger::IncludeLocation = b;
            FileLogger::IncludeLocation = b;
        };

        void IncludeColor(bool b) { ConsoleLogger::IncludeColor = b; };


    };

    class GenericLogger : public CompoundLogger {
    public:
        GenericLogger(const std::string &context,
                      std::ofstream &file,
                      Color4 contextColor = Colors::White,
                      Color4 timestampColor = Colors::Purples::Fuchsia,
                      Color4 locationColor = Colors::Pinks::Pink,
                      Color4 pointerColor = Colors::Pinks::LightPink,
                      Color4 messageColor = Colors::Greens::LightGreen)
                : CompoundLogger(context, file) {
            this->contextColor = contextColor;
            this->timestampColor = timestampColor;
            this->locationColor = locationColor;
            this->pointerColor = pointerColor;
            this->messageColor = messageColor;
        }
    };

    extern std::ofstream GlobalLogFile;

    extern GenericLogger Info;
    extern GenericLogger Warning;
    extern GenericLogger Error;
    extern GenericLogger Fatal;
    extern GenericLogger Verbose;
    extern GenericLogger Debug;

    class LibraryLogger {
    public:
        LibraryLogger(const std::string libname) :
                Info {libname + "::" + "info", GlobalLogFile, Colors::Green, Colors::Gray, Colors::Gray, Colors::Green, Colors::White},
                Warning {libname + "::" + "warning", GlobalLogFile, Colors::Yellow, Colors::Gray, Colors::Gray, Colors::Yellow, Colors::White},
                Error {libname + "::" + "error", GlobalLogFile, Colors::Red, Colors::Gray, Colors::Gray, Colors::Red, Colors::White},
                Fatal {libname + "::" + "fatal", GlobalLogFile, Colors::Reds::Crimson, Colors::Gray, Colors::Gray, Colors::Reds::Crimson, Colors::White},
                Verbose {libname + "::" + "verbose", GlobalLogFile, Colors::Blue, Colors::Gray, Colors::Gray, Colors::Blue, Colors::White},
                Debug {libname + "::" + "debug", GlobalLogFile, Colors::Purples::Purple, Colors::Gray, Colors::Gray, Colors::Purples::Purple, Colors::White}
        {}
    public:
        GenericLogger Info;// {"info", GlobalLogFile, Colors::Green, Colors::Gray, Colors::Gray, Colors::Green, Colors::White};
        GenericLogger Warning;// {"warning", GlobalLogFile, Colors::Yellow, Colors::Gray, Colors::Gray, Colors::Yellow, Colors::White};
        GenericLogger Error;// {"error", GlobalLogFile, Colors::Red, Colors::Gray, Colors::Gray, Colors::Red, Colors::White};
        GenericLogger Fatal;// {"fatal", GlobalLogFile, Colors::Reds::Crimson, Colors::Gray, Colors::Gray, Colors::Reds::Crimson, Colors::White};
        GenericLogger Verbose;// {"verbose", GlobalLogFile, Colors::Blue, Colors::Gray, Colors::Gray, Colors::Blue, Colors::White};
        GenericLogger Debug;// {"debug", GlobalLogFile, Colors::Purples::Purple, Colors::Gray, Colors::Gray, Colors::Purples::Purple, Colors::White};
    public:
        void EnableAll(bool b) {
            Info.EnableConsole(b);
            Info.EnableFile(b);
            Warning.EnableConsole(b);
            Warning.EnableFile(b);
            Error.EnableConsole(b);
            Error.EnableFile(b);
            Fatal.EnableConsole(b);
            Fatal.EnableFile(b);
            Verbose.EnableConsole(b);
            Verbose.EnableFile(b);
            Debug.EnableConsole(b);
            Debug.EnableFile(b);
        }
    };
}
